"""
Auth API.
"""
from algora.api.service.auth.asynchronous import (
    async_login, async_refresh_token, async_exchange_token
)
from algora.api.service.auth.synchronous import (
    login, refresh_token, exchange_token
)



"""
Date and time functions.
"""
import calendar
from datetime import date, datetime, time
from functools import lru_cache
from typing import List, Union

import pandas as pd
import pytz
from pandas import to_datetime, bdate_range
from pandas_market_calendars import get_calendar

from algora.common.date import DayCountConvention, SECONDS_IN_DAY
from algora.common.type import Datetime


@lru_cache(10)
def seconds_in_year(day_count_convention: DayCountConvention) -> float:
    """
    Calculate number of seconds per year based on day count convention. LRU cache is used to have 4x performance.

    Args:
        day_count_convention (DayCountConvention): Day count convention to use in calculation

    Returns:
        float: Number of seconds per year based on day count convention
    """
    return SECONDS_IN_DAY * float(day_count_convention)


def date_to_timestamp(as_of_date: date) -> int:
    """
    Convert date to epoch timestamp in milliseconds.

    Args:
        as_of_date (date): Python date

    Returns:
        int: Epoch timestamp in milliseconds
    """
    return calendar.timegm(as_of_date.timetuple()) * 1000


def datetime_to_timestamp(dt: datetime) -> int:
    """
    Convert datetime to epoch timestamp in milliseconds.

    Args:
        dt (datetime): Python datetime

    Returns:
        int: Epoch timestamp in milliseconds
    """
    return calendar.timegm(dt.utctimetuple()) * 1000


def datetime_to_utc(dt: datetime) -> datetime:
    """
    Standardize datetime to UTC. Assume that datetime where `tzinfo=None` is already in UTC.

    Args:
        dt (datetime): Python datetime

    Returns:
        datetime: Python datetime with standardized UTC timezone (`tzinfo=None`)
    """
    # assume that datetime without timezone is already in UTC
    if dt.tzinfo is None:
        return dt
    return dt.astimezone(pytz.utc).replace(tzinfo=None)


def date_to_datetime(as_of_date: date, as_of_time: time = datetime.min.time()) -> datetime:
    """
    Convert date and optional time to datetime. NOTE: time should not contain a timezone or else offset may not be
    correct.

    Args:
        as_of_date (date): Python date
        as_of_time (time): Python time

    Returns:
        datetime: Python datetime
    """
    return datetime.combine(as_of_date, as_of_time)


def standardize_date_or_datetime(as_of_date: Datetime) -> datetime:
    """
    Convert Datetime (Union[date, datetime]) to standardized UTC datetime.

    Args:
        as_of_date (date): Datetime

    Returns:
        datetime: Python UTC datetime
    """
    if isinstance(as_of_date, datetime):
        return datetime_to_utc(as_of_date)

    return date_to_datetime(as_of_date)


def timestamp_to_datetime(timestamp: int) -> datetime:
    """
    Transform epoch timestamp in milliseconds to datetime.

    Args:
        timestamp (int): Epoch timestamp in milliseconds

    Returns:
        datetime: Python datetime
    """
    return datetime.fromtimestamp(timestamp / 1000)


def delta_t(start: Union[datetime, pd.Series],
            end: Union[datetime, pd.Series],
            day_count_convention: DayCountConvention = DayCountConvention.ACTUAL_365_2425) -> Union[float, pd.Series]:
    """
    Calculate time to expiration given start series and end series, as well as numeric day_count.

    Args:
        start (datetime): Start (report) datetime
        end (datetime): End datetime, e.g. expiration datetime
        day_count_convention (DayCountConvention): Day count convention. Defaults to ACTUAL/365.2425

    Returns:
        float: Delta in time
    """
    denominator = seconds_in_year(day_count_convention)

    start_utc = datetime_to_utc(start) if isinstance(start, datetime) else start
    end_utc = datetime_to_utc(end) if isinstance(end, datetime) else end

    if isinstance(start_utc, pd.Series) or isinstance(end_utc, pd.Series):
        return (end_utc - start_utc).dt.total_seconds() / denominator
    else:
        return (end_utc - start_utc).total_seconds() / denominator


def create_dates_between(start: date, end: date, frequency: str = 'B') -> List[date]:
    """
    Create dates between start and end date (inclusive). Frequency used to determine which days of the week are used.

    Args:
        start (date): Python date
        end (date): Python date
        frequency (str): Frequency for date range

    Returns:
        List[date]: List of dates
    """
    return [dt.date() for dt in to_datetime(bdate_range(start=start, end=end, freq=frequency).to_list())]


def create_market_dates_between(start: date, end: date, name: str = 'NYSE') -> List[date]:
    """
    Create dates between start and end date (inclusive) given exchange calendar.

    Args:
        start (date): Python date
        end (date): Python date
        name (str): Calendar name, such as 'NYSE'

    Returns:
        List[date]: List of dates
    """
    return [dt.date() for dt in
            to_datetime(get_calendar(name).schedule(start_date=start, end_date=end).index).to_list()]



"""
Execution event API.
"""
from algora.api.service.execution_event.asynchronous import (
    async_get_all_execution_events, async_get_all_portfolio_states, async_get_all_cash_payments,
    async_create_execution_event
)
from algora.api.service.execution_event.synchronous import (
    get_all_execution_events, get_all_portfolio_states, get_all_cash_payments, create_execution_event
)



"""
Project Package API
"""
from algora.api.service.project.package.synchronous import (
    get_project_packages, get_project_package, create_project_package, update_project_package, delete_project_package,
    search_project_packages
)
from algora.api.service.project.package.asynchronous import (
    async_get_project_packages, async_get_project_package, async_create_project_package,
    async_update_project_package, async_delete_project_package, async_search_project_packages
)



"""
Data request cache implementation. Data requests are stored as a key in the cache using format:
{timestamp}.{request hash}. Cache supports single and start/end date requests.

Cache 'put' cases:
    date:
    1. Query data
    2. Create key with single date and hashed request
    3. Store in cache

    start_date and end_date:
    1. Query data.
    2. Parse response to get all individuals dates (assume column is 'date', but allow column to be overwritten).
       Currently only supports DataFrame
    3. Group the response by date and store in cache
"""
from datetime import date
from functools import partial
from typing import Dict, Tuple

import pandas as pd

from algora.common.decorators import partitioned_cached
from algora.common.date import date_to_timestamp, create_dates_between
from algora.common.model import DataRequest


def __get_request(*args, **kwargs):
    for arg in args:
        if isinstance(arg, DataRequest):
            return arg

    for arg in kwargs.values():
        if isinstance(arg, DataRequest):
            return arg


def __partition_key_mapping(request: DataRequest) -> Tuple[date, str]:
    if isinstance(request.as_of_date, date):
        timestamp = date_to_timestamp(request.as_of_date)
        return request.as_of_date, build_partition_key(timestamp, request)
    else:
        timestamp = date_to_timestamp(request.as_of_date.date())
        return request.as_of_date.date(), build_partition_key(timestamp, request)


def build_partition_key(timestamp: int, request: DataRequest) -> str:
    excluded_keys = {'as_of_date', 'start_date', 'end_date', 'as_of'}
    return f"{timestamp}.{request.base64_encoded(exclude=excluded_keys)}"


def get_key_partitions(request: DataRequest) -> Dict[date, str]:
    partitions = []
    if request.as_of_date:
        partitions.append(__partition_key_mapping(request))
    elif request.start_date and request.end_date:
        start = request.start_date.date()
        end = request.end_date.date()

        for _date in create_dates_between(start=start, end=end, frequency='D'):
            base_request = request.copy(deep=True)
            base_request.as_of_date = _date
            base_request.start_date = None
            base_request.end_date = None
            partitions.append(__partition_key_mapping(base_request))

    return dict(partitions)


day_partition_cache = partial(
    partitioned_cached,
    get_key_partitions=get_key_partitions,
    combine_partitions=pd.concat,
    partition_value=lambda df: {dt.date(): data for dt, data in df.groupby(pd.Grouper(key='date', freq='D'))},
    cache_lookup_criterion=lambda key: key.as_of_date is not None,
    key=__get_request
)



from algora.api.service.runner.model import RunnerRequest


def _get_runners_request_info() -> dict:
    return {
        'endpoint': f"config/runner"
    }


def _get_runner_request_info(id: str) -> dict:
    return {
        'endpoint': f"config/runner{id}"
    }


def _create_runner_request_info(request: RunnerRequest) -> dict:
    return {
        'endpoint': f"config/runner",
        'json': request.request_dict()
    }


def _update_runner_request_info(id: str, request: RunnerRequest) -> dict:
    return {
        'endpoint': f"config/runner/{id}",
        'json': request.request_dict()
    }


def _delete_runner_request_info(id: str) -> dict:
    return {
        'endpoint': f"config/runner/{id}"
    }



from typing import Optional, Dict

from algora.api.service.project.enum import ProjectType
from algora.common.base import Base


class ProjectRequest(Base):
    name: str
    description: Optional[str]
    language: str
    entrypoint: str
    args: Dict[str, str]
    type: ProjectType
    packages: Dict[str, str]


class ProjectResponse(Base):
    id: str
    name: str
    description: Optional[str]
    language: str
    entrypoint: str
    args: Dict[str, str]
    type: ProjectType
    packages: Dict[str, str]
    created_by: str
    created_at: int
    updated_by: Optional[str]
    updated_at: Optional[int]



from typing import Dict, Any, List

from algora.api.service.runner.__util import (
    _create_runner_request_info, _get_runner_request_info, _get_runners_request_info, _update_runner_request_info,
    _delete_runner_request_info
)
from algora.api.service.runner.model import RunnerRequest
from algora.common.function import no_transform
from algora.common.decorators import async_data_request
from algora.common.requests import (
    __async_put_request, __async_get_request, __async_post_request, __async_delete_request
)


@async_data_request(transformer=no_transform)
async def async_get_runner(id: str) -> Dict[str, Any]:
    """
    Asynchronously get runner by ID.

    Args:
        id (str): Runner ID

    Returns:
        Dict[str, Any]: Runner response
    """
    request_info = _get_runner_request_info(id)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_get_runners() -> List[Dict[str, Any]]:
    """
    Asynchronously get all runners.

    Returns:
        List[Dict[str, Any]]: List of Runner response
    """
    request_info = _get_runners_request_info()
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_create_runner(request: RunnerRequest) -> Dict[str, Any]:
    """
    Asynchronously get or create runner.

    Returns:
        Dict[str, Any]: Runner response
    """
    request_info = _create_runner_request_info(request)
    return await __async_put_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_update_runner(id: str, request: RunnerRequest) -> Dict[str, Any]:
    """
    Asynchronously update runner.

    Args:
        id (str): Runner ID
        request (RunnerRequest): Runner request

    Returns:
        Dict[str, Any]: Runner response
    """
    request_info = _update_runner_request_info(id, request)
    return await __async_post_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_delete_runner(id: str) -> None:
    """
    Asynchronously delete runner by ID.

    Args:
        id (str): Runner ID

    Returns:
        None
    """
    request_info = _delete_runner_request_info(id)
    return await __async_delete_request(**request_info)




"""
Project API.
"""
from algora.api.service.project.synchronous import (
    get_project, get_projects, create_project, update_project, copy_project, delete_project
)
from algora.api.service.project.asynchronous import (
    async_get_project, async_get_projects, async_create_project, async_update_project, async_copy_project,
    async_delete_project
)



from typing import Dict, Any, List

from algora.api.service.project.__util import (
    _create_project_request_info, _get_project_request_info, _get_projects_request_info, _update_project_request_info,
    _delete_project_request_info, _copy_project_request_info
)
from algora.api.service.project.model import ProjectRequest
from algora.common.decorators import async_data_request
from algora.common.function import no_transform
from algora.common.requests import (
    __async_put_request, __async_get_request, __async_post_request, __async_delete_request
)


@async_data_request(transformer=no_transform)
async def async_get_project(id: str) -> Dict[str, Any]:
    """
    Asynchronously get project by ID.
    
    Args:
        id (str): Project ID

    Returns:
        Dict[str, Any]: Project response
    """
    request_info = _get_project_request_info(id)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_get_projects() -> List[Dict[str, Any]]:
    """
    Asynchronously get all projects.

    Returns:
        List[Dict[str, Any]]: List of project response
    """
    request_info = _get_projects_request_info()
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_create_project(request: ProjectRequest) -> Dict[str, Any]:
    """
    Asynchronously create project.

    Args:
        request (ProjectRequest): Project request

    Returns:
        Dict[str, Any]: Project response
    """
    request_info = _create_project_request_info(request)
    return await __async_put_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_update_project(id: str, request: ProjectRequest) -> Dict[str, Any]:
    """
    Asynchronously update project.

    Args:
        id (str): Project ID
        request (ProjectRequest): Project request

    Returns:
        Dict[str, Any]: Project response
    """
    request_info = _update_project_request_info(id, request)
    return await __async_post_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_copy_project(id: str) -> Dict[str, Any]:
    """
    Asynchronously update project.

    Args:
        id (str): Project ID

    Returns:
        Dict[str, Any]: Project response
    """
    request_info = _copy_project_request_info(id)
    return await __async_post_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_delete_project(id: str) -> None:
    """
    Asynchronously delete project by ID.

    Args:
        id (str): Project ID

    Returns:
        None
    """
    request_info = _delete_project_request_info(id)
    return await __async_delete_request(**request_info)



from typing import Optional

from algora.api.service.project.package.enum import ProjectPackageTag
from algora.common.base import Base


class ProjectPackageRequest(Base):
    name: str
    version: str
    tag: Optional[ProjectPackageTag]
    path: Optional[str]


class ProjectPackageResponse(Base):
    id: str
    name: str
    version: str
    tag: Optional[ProjectPackageTag]
    path: Optional[str]
    created_by: str
    created_at: int
    updated_by: Optional[str]
    updated_at: Optional[int]



"""
Enums for payment frequency, day count convention and period.
"""
from algora.common.base_enum import BaseEnum


class PaymentFrequency(BaseEnum):
    """
    Payment frequency used to discount cashflows and accrue interest.
    """

    DAILY = "DAILY"
    WEEKLY = "WEEKLY"
    SEMI_MONTHLY = "SEMI_MONTHLY"
    MONTHLY = "MONTHLY"
    SEMI_QUARTERLY = "SEMI_QUARTERLY"
    QUARTERLY = "QUARTERLY"
    TRI_ANNUALLY = "TRI_ANNUALLY"
    SEMI_ANNUALLY = "SEMI_ANNUALLY"
    ANNUALLY = "ANNUALLY"

    def __int__(self):
        if self == PaymentFrequency.DAILY:
            return 252
        elif self == PaymentFrequency.WEEKLY:
            return 52
        elif self == PaymentFrequency.SEMI_MONTHLY:
            return 26
        elif self == PaymentFrequency.MONTHLY:
            return 12
        elif self == PaymentFrequency.SEMI_QUARTERLY:
            return 6
        elif self == PaymentFrequency.QUARTERLY:
            return 4
        elif self == PaymentFrequency.TRI_ANNUALLY:
            return 3
        elif self == PaymentFrequency.SEMI_ANNUALLY:
            return 2
        elif self == PaymentFrequency.ANNUALLY:
            return 1
        else:
            raise Exception(f"Payment frequency type {self} not supported")


class PeriodsPerYear(BaseEnum):
    """
    Periods per year. TODO deprecate
    """
    DAILY = "DAILY"
    WEEKLY = "WEEKLY"
    MONTHLY = "MONTHLY"
    ANNUALLY = "ANNUALLY"

    def __int__(self):
        if self == PeriodsPerYear.DAILY:
            return 365
        elif self == PeriodsPerYear.MONTHLY:
            return 12
        else:
            raise Exception(f"PeriodsPerYear type {self} not supported")


class DayCountConvention(BaseEnum):
    """
    Day count convention to determine how interest accrues over payment periods.
    """

    # Actual/360: Number of days between dates divided by 360
    ACTUAL_360 = "ACTUAL_360"

    # Actual/364: Number of days between dates divided by 364
    ACTUAL_364 = "ACTUAL_364"

    # Actual/365 FIXED: Number of days between dates divided by 365
    ACTUAL_365F = "ACTUAL_365F"

    # Actual/365_2425: Number of days between dates divided by 365.25
    ACTUAL_365_2425 = "ACTUAL_365_2425"

    def __float__(self):
        if self == DayCountConvention.ACTUAL_360:
            return 360
        elif self == DayCountConvention.ACTUAL_364:
            return 364
        elif self == DayCountConvention.ACTUAL_365F:
            return 365
        elif self == DayCountConvention.ACTUAL_365_2425:
            return 365.2425
        else:
            raise Exception(f"Day count convention type {self} not supported")

    def __int__(self):
        return int(float(self))


class TimeUnit(BaseEnum):
    """
    Time unit.
    """

    DAY = "DAY"
    WEEK = "WEEK"
    MONTH = "MONTH"
    YEAR = "YEAR"


SECONDS_IN_MINUTE = 60
MINUTES_IN_HOUR = 60
HOURS_IN_DAY = 24
DAYS_IN_WEEK = 7

SECONDS_IN_DAY = SECONDS_IN_MINUTE * MINUTES_IN_HOUR * HOURS_IN_DAY



from typing import Any

from algora.api.service.project.package.model import ProjectPackageRequest


def _get_project_package_request_info(id: str) -> dict:
    return {
        'endpoint': f"config/project/package/{id}"
    }


def _get_project_packages_request_info() -> dict:
    return {
        'endpoint': f"config/project/package"
    }


def _search_project_packages_request_info(name: str, version: str) -> dict:
    return {
        'endpoint': f"config/project/package/search?name={name}&version={version}"
    }


def _create_project_package_request_info(request: ProjectPackageRequest) -> dict:
    return {
        'endpoint': f"config/project/package",
        'json': request.request_dict()
    }


def _update_project_package_request_info(id: str, request: ProjectPackageRequest) -> dict:
    return {
        'endpoint': f"config/project/package/{id}",
        'json': request.request_dict()
    }


def _delete_project_package_request_info(id: str) -> dict:
    return {
        'endpoint': f"config/project/package/{id}"
    }



"""
Helper functions.
"""
from functools import reduce
from typing import Dict, Any, Union, List, Optional

import pandas as pd
from aiohttp import ClientResponse
from pandas import DataFrame
from requests import Response


def coalesce(*args):
    """
    Coalesce operator returns the first arg that isn't None.

    Args:
        *args: Tuple of args passed to the function

    Returns:
        The first non-None value in *args
    """
    return reduce(lambda x, y: x if x is not None else y, args)


def coalesce_callables(*args):
    """
    Coalesce operator returns the first arg that isn't None. If the arg is callable it will check that the value
    returned from calling the arg is not none and then return the value from calling it.

    WARNING: If an argument implements __call__ this method will evaluate the return of the __call__ method and return
    that instead of the argument itself. This is important when using python classes.

    Args:
        *args: Tuple of args passed to the function

    Returns:
        The first non-None value in *args
    """
    for arg in args:
        value = arg() if callable(arg) else arg
        if value is not None:
            return value
    return None


def transform_one_or_many(
        data: Union[List[Dict[str, Any]], Dict[str, List[Dict[str, Any]]]],
        key: Optional[str] = None
) -> Union[DataFrame, Dict[str, DataFrame]]:
    """
    Converts data to a dataframe or multiple dataframes.

    Args:
        data (Union[List[Dict[str, Any]], Dict[str, List[Dict[str, Any]]]]): Dict being transformed into Dataframes
        key (Optional[str]): Key for indexing the values of the data dict

    Returns:
        Union[DataFrame, Dict[str, DataFrame]]: A single dataframe or a map of names to dataframes
    """
    if isinstance(data, dict):
        for k in data.keys():
            data[k] = DataFrame(data[k][key])
        return data

    return DataFrame(data)


def to_pandas_with_index(data: Dict[str, Any], index: str = 'date') -> DataFrame:
    """
    Transform input data into pandas dataframe and assign index.

    Args:
        data (Dict[str, Any]): Input data as dict
        index (str): Index column name

    Returns:
        DataFrame: DataFrame
    """
    # necessary to drop column in order to avoid duplicates when converting to json
    return pd.DataFrame(data).set_index(index, drop=True)


def no_transform(data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Passthrough data without any transformation.

    Args:
        data (Dict[str, Any]): Input data

    Returns:
        Dict[str, Any]: Output data
    """
    return data


async def build_response_obj(aio_response: ClientResponse) -> Response:
    """
    Construct a requests Response object from an aiohttp Client Response object

    Args:
        aio_response (ClientResponse): Response object from aiohttp

    Returns:
        Response: A requests Response object
    """
    content = await aio_response.content.read()

    response = Response()
    response._content = content
    response.url = str(aio_response.url)
    response.status_code = aio_response.status
    headers = {row[0]: row[1] for row in aio_response.headers.items()}
    response.headers = headers
    return response



import functools
from typing import TypeVar, Callable, Dict, List

KEY = TypeVar("KEY")
VALUE = TypeVar("VALUE")
PARTITION_ID = TypeVar("PARTITION_ID")


def partitioned_cached(
        cache,
        *,
        get_key_partitions: Callable[[KEY], Dict[PARTITION_ID, str]],
        combine_partitions: Callable[[List[VALUE]], VALUE],
        partition_value: Callable[[VALUE], Dict[PARTITION_ID, VALUE]],
        cache_lookup_criterion: Callable[[KEY], bool] = lambda key: True,
        key=lambda x: x,
        lock=None
):
    """
    Decorator to wrap a function with a memoizing callable that saves results in a cache.
    """

    def decorator(func):
        if cache is None:

            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            def clear():
                pass

        elif lock is None:

            def wrapper(*args, **kwargs):
                k = key(*args, **kwargs)
                partitions = get_key_partitions(k)
                if cache_lookup_criterion(k):
                    try:
                        partition_values = [cache[cache_key] for cache_key in partitions.values()]
                        return combine_partitions(partition_values)
                    except KeyError:
                        pass  # key not found
                v = func(*args, **kwargs)
                partitioned_value = partition_value(v)
                try:
                    for partition_key_identifier, value in partitioned_value.items():
                        k = partitions[partition_key_identifier]
                        cache[k] = value
                except ValueError:
                    pass  # value too large
                return v

            def clear():
                cache.clear()

        else:

            def wrapper(*args, **kwargs):
                k = key(*args, **kwargs)
                partitions = get_key_partitions(k)
                if cache_lookup_criterion(k):
                    try:
                        with lock:
                            partition_values = [cache[cache_key] for cache_key in partitions.values()]
                            return combine_partitions(partition_values)
                    except KeyError:
                        pass  # key not found
                v = func(*args, **kwargs)
                partitioned_value = partition_value(v)
                # in case of a race, prefer the item already in the cache
                try:
                    with lock:
                        for partition_key_identifier, value in partitioned_value.items():
                            k = partitions[partition_key_identifier]
                            cache[k] = value
                except ValueError:
                    return v  # value too large

            def clear():
                with lock:
                    cache.clear()

        wrapper.cache = cache
        wrapper.cache_key = key
        wrapper.cache_lock = lock
        wrapper.cache_clear = clear

        return functools.update_wrapper(wrapper, func)

    return decorator



from typing import Dict, Any, List

from algora.api.service.runner.__util import (
    _create_runner_request_info, _get_runner_request_info, _get_runners_request_info, _update_runner_request_info,
    _delete_runner_request_info
)
from algora.api.service.runner.model import RunnerRequest
from algora.common.decorators import data_request
from algora.common.function import no_transform
from algora.common.requests import __put_request, __get_request, __post_request, __delete_request


@data_request(transformer=no_transform)
def get_runner(id: str) -> Dict[str, Any]:
    """
    Get runner by ID.

    Returns:
        Dict[str, Any]: Runner response
    """
    request_info = _get_runner_request_info(id)
    return __get_request(**request_info)


@data_request(transformer=no_transform)
def get_runners() -> List[Dict[str, Any]]:
    """
    Get all runners.

    Returns:
        List[Dict[str, Any]]: List of runner response
    """
    request_info = _get_runners_request_info()
    return __get_request(**request_info)


@data_request(transformer=no_transform)
def create_runner(request: RunnerRequest) -> Dict[str, Any]:
    """
    Get or create runner.

    Returns:
        Dict[str, Any]: Runner response
    """
    request_info = _create_runner_request_info(request)
    return __put_request(**request_info)


@data_request(transformer=no_transform)
def update_runner(id: str, request: RunnerRequest) -> Dict[str, Any]:
    """
    Update runner.

    Args:
        id (str): Runner ID
        request (RunnerRequest): Runner request

    Returns:
        Dict[str, Any]: Runner response
    """
    request_info = _update_runner_request_info(id, request)
    return __post_request(**request_info)


@data_request(transformer=no_transform)
def delete_runner(id: str) -> None:
    """
    Delete Runner by ID.

    Args:
        id (str): Runner ID

    Returns:
        None
    """
    request_info = _delete_runner_request_info(id)
    return __delete_request(**request_info)



from typing import Dict, List


from algora.api.service.resource.__util import _upload_resource_request_info, _delete_resource_request_info
from algora.common.decorators import async_data_request
from algora.common.function import no_transform
from algora.common.requests import __async_put_request, __async_delete_request


@async_data_request(
    transformer=no_transform,
    processor=lambda response: response.status_code
)
async def async_upload_resources(path: str, file_locations: List[str]) -> None:
    """
    Asynchronously uploads the files.

    Args:
        path (str): Path where data is uploaded to
        file_locations (List[str]): List of paths to the file being uploaded

    Returns:
        Dict[str, Any]: Project package response
    """
    # aiofiles doesnt work atm so we are using the blocking open
    files = [open(location, 'rb') for location in file_locations]
    request_info = _upload_resource_request_info(path, files)
    response = await __async_put_request(**request_info)
    [file.close() for file in files]
    return response


@async_data_request(
    transformer=no_transform,
    processor=lambda response: response.status_code
)
async def async_delete_resource(path: str) -> None:
    """
    Asynchronously deletes the wheel file for the project package.

    Args:
        path (str): Path where data is uploaded to

    Returns:
        Dict[str, Any]: Project package response
    """
    request_info = _delete_resource_request_info(path)
    return await __async_delete_request(**request_info)



from algora.common.date.enum import PaymentFrequency, DayCountConvention, PeriodsPerYear, TimeUnit, SECONDS_IN_MINUTE, \
    MINUTES_IN_HOUR, HOURS_IN_DAY, DAYS_IN_WEEK, SECONDS_IN_DAY
from algora.common.date.function import seconds_in_year, date_to_timestamp, datetime_to_timestamp, datetime_to_utc, \
    date_to_datetime, standardize_date_or_datetime, timestamp_to_datetime, delta_t, create_dates_between, \
    create_market_dates_between



"""
Common models.
"""
from datetime import datetime
from typing import Optional

from pydantic import Field as pyField

from algora.common.base import Base
from algora.common.enum import FieldType
from algora.common.type import Datetime


class Field(Base):
    logical_name: str
    type: FieldType


class DataRequest(Base):
    """
    Base data request, inherited by all instrument-specific data request classes.
    """
    as_of_date: Optional[Datetime] = pyField(default=None)
    start_date: Optional[datetime] = pyField(default=None)
    end_date: Optional[datetime] = pyField(default=None)
    as_of: Optional[datetime] = pyField(default=None)



from typing import Optional

from algora.api.service.runner.enum import RunnerType
from algora.common.base import Base


class RunnerRequest(Base):
    name: str
    type: RunnerType
    min_cpu: str
    max_cpu: str
    min_ram: str
    max_ram: str
    image: Optional[str]


class RunnerResponse(Base):
    id: str
    name: str
    type: RunnerType
    min_cpu: str
    max_cpu: str
    min_ram: str
    max_ram: str
    image: Optional[str]
    created_by: str
    created_at: int
    updated_by: Optional[str]
    updated_at: Optional[int]



from algora.api.service.project.model import ProjectRequest
from algora.common.enum import UpdateRequestAction


def _get_project_request_info(id: str) -> dict:
    return {
        'endpoint': f"config/project/{id}"
    }


def _get_projects_request_info() -> dict:
    return {
        'endpoint': f"config/project"
    }


def _create_project_request_info(request: ProjectRequest) -> dict:
    return {
        'endpoint': f"config/project",
        'json': request.request_dict()
    }


def _update_project_request_info(id: str, request: ProjectRequest) -> dict:
    return {
        'endpoint': f"config/project/{id}?action={UpdateRequestAction.UPDATE}",
        'json': request.request_dict()
    }


def _copy_project_request_info(id: str) -> dict:
    return {
        'endpoint': f"config/project/{id}?action={UpdateRequestAction.COPY}",
    }


def _delete_project_request_info(id: str) -> dict:
    return {
        'endpoint': f"config/project/{id}"
    }



from typing import Dict, Any

import pandas as pd

from algora.api.service.execution_event.__util import (
    _get_all_portfolio_states_request_info, _get_all_cash_payments_request_info, _create_execution_event_request_info,
    _get_all_execution_events_request_info
)
from algora.api.service.execution_event.model import ExecutionEventRequest
from algora.common.decorators import async_data_request
from algora.common.enum import Order
from algora.common.function import no_transform
from algora.common.requests import __async_put_request, __async_get_request


@async_data_request
async def async_get_all_execution_events(execution_id: str, order: Order = Order.ASC) -> pd.DataFrame:
    """
    Asynchronously get all execution events.

    Args:
        execution_id (str): Execution ID
        order (Order): SQL Order

    Returns:
        DataFrame: DataFrame of execution events
    """
    request_info = _get_all_execution_events_request_info(execution_id, order)
    return await __async_get_request(**request_info)


@async_data_request(transformer=lambda data: pd.DataFrame([r['data'] for r in data]))
async def async_get_all_portfolio_states(execution_id: str, order: Order = Order.ASC) -> pd.DataFrame:
    """
    Asynchronously get all portfolio states.

    Args:
        execution_id (str): Execution ID
        order (Order): SQL Order

    Returns:
        DataFrame: DataFrame of portfolio states
    """
    request_info = _get_all_portfolio_states_request_info(execution_id, order)
    return await __async_get_request(**request_info)


@async_data_request(transformer=lambda data: pd.DataFrame([r['data'] for r in data]))
async def async_get_all_cash_payments(execution_id: str, order: Order = Order.ASC) -> pd.DataFrame:
    """
    Asynchronously get all cash payments.

    Args:
        execution_id (str): Execution ID
        order (Order): SQL Order

    Returns:
        DataFrame: DataFrame of cash payments
    """
    request_info = _get_all_cash_payments_request_info(execution_id, order)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_create_execution_event(request: ExecutionEventRequest) -> Dict[str, Any]:
    """
    Asynchronously create execution event.

    Args:
        request (ExecutionEventRequest): Execution event request

    Returns:
        Dict[str, Any]: Execution event response
    """
    request_info = _create_execution_event_request_info(request)
    return await __async_put_request(**request_info)



from algora.common.base_enum import BaseEnum


class ProjectType(BaseEnum):
    RESEARCH_TEMPLATE = "RESEARCH_TEMPLATE"
    STRATEGY_TEMPLATE = "STRATEGY_TEMPLATE"
    RESEARCH = "RESEARCH"
    STRATEGY = "STRATEGY"
    UNKNOWN = "UNKNOWN"



from datetime import date
from typing import Optional, Dict, Any

from algora.api.service.execution_event.enum import EventType
from algora.common.base import Base


class BaseExecutionEvent(Base):
    execution_id: str
    event_type: EventType


class ExecutionEventRequest(BaseExecutionEvent):
    execution_id: str
    event_type: EventType
    data: dict


class ExecutionEventResponse(BaseExecutionEvent):
    id: str
    execution_id: str
    event_type: EventType
    data: Any
    created_at: int


class CashPaymentEvent(Base):
    payment_date: date
    type: str
    amount: float
    position_id: Optional[str]
    currency: Optional[str]


class PortfolioStateEvent(Base):
    date: date
    valuation: float
    pnl: float
    portfolio: Dict[str, Any]



from typing import Callable, Union
from pydantic import Field

from algora.common.base import Base
from algora.common.decorators.serializable.__util import parse_obj, init_subclass, get_validators, convert_to_real_type


def serializable(
        _cls: Base = None
) -> Union[Callable, Base]:
    """
    This decorator adds an identifier to the decorated class for identification purposes when serializing and
    deserializing.

    NOTE: This decorator should be used on all children of base abstract classes that use the @serializable_base_class
    decorator.

    Args:
        _cls (object): The class being decorated

    Returns:
        cls: The updated class with the descriptor field
    """

    def wrap(cls):
        return type(cls.cls_name(), (cls,), {'descriptor': Field(cls.cls_name())})

    if _cls is None:
        return wrap

    return wrap(_cls)


def serializable_base_class(
        _cls: Base = None
) -> Union[Callable, Base]:
    """
    This is a decorator sets up functionality to track all subtypes of a parent class. If you want to use a base
    class as a type signature and want the subtypes to be properly deserialized you need to add this decorator to
    the base class and @serializable to all of its children.

    NOTE: You should only use this decorator for base abstract classes

    Args:
        _cls (object): The class being decorated

    Returns:
        cls: The updated class with the descriptor field
    """

    def wrap(cls):
        setattr(cls, '_subtypes_', dict())
        setattr(cls, 'parse_obj', classmethod(parse_obj))
        setattr(cls, '__init_subclass__', classmethod(init_subclass))
        setattr(cls, '__get_validators__', classmethod(get_validators))
        setattr(cls, '_convert_to_real_type_', classmethod(convert_to_real_type))
        return cls

    if _cls is None:
        return wrap

    return wrap(_cls)



from typing import Dict, Any, List

from algora.api.service.project.package.__util import (
    _create_project_package_request_info, _get_project_package_request_info, _get_project_packages_request_info,
    _update_project_package_request_info, _delete_project_package_request_info, _search_project_packages_request_info
)
from algora.api.service.project.package.model import ProjectPackageRequest
from algora.common.decorators import async_data_request
from algora.common.function import no_transform
from algora.common.requests import (
    __async_put_request, __async_get_request, __async_post_request, __async_delete_request
)


@async_data_request(transformer=no_transform)
async def async_get_project_package(id: str) -> Dict[str, Any]:
    """
    Asynchronously get project package by ID.
    
    Args:
        id (str): Project package ID

    Returns:
        Dict[str, Any]: Project package response
    """
    request_info = _get_project_package_request_info(id)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_get_project_packages() -> List[Dict[str, Any]]:
    """
    Asynchronously get all project packages.

    Returns:
        List[Dict[str, Any]]: List of project package response
    """
    request_info = _get_project_packages_request_info()
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_search_project_packages(name: str, version: str) -> List[Dict[str, Any]]:
    """
    Asynchronously search for a project package based off of the name and version.

    Parameters:
        name: Project package name
        version: Project package version

    Returns:
        Dict[str, Any]: Project package response
    """
    request_info = _search_project_packages_request_info(name, version)
    return await __async_get_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_create_project_package(request: ProjectPackageRequest) -> Dict[str, Any]:
    """
    Asynchronously create project package.

    Args:
        request (ProjectPackageRequest): Project package request

    Returns:
        Dict[str, Any]: Project package response
    """
    request_info = _create_project_package_request_info(request)
    return await __async_put_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_update_project_package(id: str, request: ProjectPackageRequest) -> Dict[str, Any]:
    """
    Asynchronously update project package.

    Args:
        id (str): Project package ID
        request (ProjectPackageRequest): Project package request

    Returns:
        Dict[str, Any]: Project package response
    """
    request_info = _update_project_package_request_info(id, request)
    return await __async_post_request(**request_info)


@async_data_request(transformer=no_transform)
async def async_delete_project_package(id: str) -> None:
    """
    Asynchronously delete project package by ID.

    Args:
        id (str): Project package ID

    Returns:
        None
    """
    request_info = _delete_project_package_request_info(id)
    return await __async_delete_request(**request_info)



from algora.api.service.execution_event.model import ExecutionEventRequest
from algora.common.enum import Order


def _get_all_execution_events_request_info(execution_id: str, order: Order = Order.ASC) -> dict:
    config = {
        "endpoint": f"config/execution/event/{execution_id}?order={order}"
    }

    return config


def _get_all_portfolio_states_request_info(execution_id: str, order: Order = Order.ASC) -> dict:
    config = {
        "endpoint": f"config/execution/event/{execution_id}?type=PORTFOLIO&order={order}"
    }

    return config


def _get_all_cash_payments_request_info(execution_id: str, order: Order = Order.ASC) -> dict:
    config = {
        "endpoint": f"config/execution/event/{execution_id}?type=CASH_PAYMENT&order={order}"
    }

    return config


def _create_execution_event_request_info(request: ExecutionEventRequest) -> dict:
    return {
        "endpoint": "config/execution/event",
        "json": request.request_dict()
    }



from typing import List, Dict, Optional


def _get_client_id_request_info() -> dict:
    return {
        "endpoint": "clients?clientId=realm-management",
        "url_key": "keycloak_admin"
    }


def _get_users_request_info(limit: int, page_offset: int) -> dict:
    return {
        "endpoint": "users",
        "params": {"max": limit, "first": page_offset},
        "url_key": "keycloak_admin"
    }


def _get_user_count_request_info() -> dict:
    return {
        "endpoint": f"users/count",
        "url_key": "keycloak_admin"
    }


def _get_user_request_info(user_id: str) -> dict:
    return {
        "endpoint": f"users/{user_id}",
        "url_key": "keycloak_admin"
    }


def _execute_actions_email_request_info(user_id: str, actions: List[str], client_id: str, redirect_uri: str) -> dict:
    return {
        "endpoint": f"users/{user_id}/execute-actions-email",
        "params": {"client_id": client_id, "redirect_uri": redirect_uri, "lifespan": 604800},
        "url_key": "keycloak_admin",
        "json": actions
    }


def _update_user_request_info(user_id: str, data: dict) -> dict:
    return {
        "endpoint": f"users/{user_id}",
        "url_key": "keycloak_admin",
        "json": data
    }


def _get_groups_request_info() -> dict:
    return {
        "endpoint": "groups",
        "url_key": "keycloak_admin"
    }


def _get_group_members_request_info(id: str) -> dict:
    return {
        "endpoint": f"groups/{id}/members",
        "url_key": "keycloak_admin"
    }


def _add_group_to_user_request_info(user_id: str, group_id: str) -> dict:
    return {
        "endpoint": f"users/{user_id}/groups/{group_id}",
        "url_key": "keycloak_admin"
    }


def _delete_group_from_user_request_info(user_id: str, group_id: str) -> dict:
    return {
        "endpoint": f"users/{user_id}/groups/{group_id}",
        "url_key": "keycloak_admin"
    }


def _get_roles_request_info() -> dict:
    return {
        "endpoint": "roles",
        "url_key": "keycloak_admin"
    }


def _get_roles_for_user_request_info(user_id: str) -> dict:
    return {
        "endpoint": f"users/{user_id}/role-mappings",
        "url_key": "keycloak_admin"
    }


def _add_role_mapping_to_user_request_info(user_id: str, role_ids: List[Dict[str, str]]) -> dict:
    return {
        "endpoint": f"users/{user_id}/role-mappings/realm",
        "url_key": "keycloak_admin",
        "json": role_ids
    }


def _get_offline_token_request_info(user_id: str) -> dict:
    return {
        "endpoint": f"config/offline_token/{user_id}"
    }


def _create_offline_token_request_info(user_id: str, refresh_token: str, exp: Optional[int] = None) -> dict:
    return {
        "endpoint": f"config/offline_token",
        "json": {
            "user_id": user_id,
            "refresh_token": refresh_token,
            "exp": exp
        }
    }


def _update_offline_token_request_info(user_id: str, refresh_token: str, exp: Optional[int] = None) -> dict:
    return {
        "endpoint": f"config/offline_token/{user_id}",
        "json": {
            "user_id": user_id,
            "refresh_token": refresh_token,
            "exp": exp
        }
    }


def _delete_offline_token_request_info(user_id: str) -> dict:
    return {
        "endpoint": f"config/offline_token/{user_id}"
    }



"""
Runner API
"""
from algora.api.service.runner.synchronous import (
    get_runner, get_runners, create_runner, update_runner, delete_runner
)
from algora.api.service.runner.asynchronous import (
    async_get_runner, async_get_runners, async_create_runner, async_update_runner, async_delete_runner
)



"""
Base classes with custom attributes for updating, serializing and deserializing data classes and enums.
"""
import base64
import json
from abc import ABC
from datetime import date, datetime, tzinfo

from pydantic import BaseModel

from algora.common.date import date_to_timestamp, datetime_to_timestamp


class Base(ABC, BaseModel):
    """
    Base class used for all data classes.
    """

    class Config:
        # use enum values when using .dict() on object
        use_enum_values = True

        json_encoders = {
            date: date_to_timestamp,
            datetime: datetime_to_timestamp,
            tzinfo: str
        }

    @classmethod
    def cls_name(cls) -> str:
        """
        Get class name.

        Returns:
            str: Class name
        """
        return cls.__name__

    def request_dict(self) -> dict:
        """
        Convert data class to dict. Used instead of `.dict()` to serialize dates as timestamps.

        Returns:
            dict: Serialized data class as dict
        """
        return json.loads(self.json())

    def base64_encoded(self, exclude=None) -> bytes:
        """
        Base-64 encode data class.

        Returns:
            bytes: Base-64 encoded data class as bytes
        """
        json_str = json.dumps(self.json(exclude=exclude), sort_keys=True)
        bytes_rep = bytes(json_str, 'utf-8')
        return base64.b64encode(bytes_rep)



from typing import Any


def _upload_resource_request_info(path: str, files: Any) -> dict:
    return {
        'endpoint': f"config/resource?path={path}",
        'files':  [('files', file) for file in files]
    }


def _delete_resource_request_info(path: str) -> dict:
    return {
        'endpoint': f"config/resource?path={path}",
    }
